"""
This module defines an UFL transformation, that takes a UFL expression
and transforms it into a sum of accumulation terms.
"""
from dune.perftool.ufl.modified_terminals import extract_modified_arguments
from dune.perftool.ufl.transformations import ufl_transformation
from dune.perftool.ufl.transformations.replace import replace_expression
from dune.perftool.ufl.transformations.identitypropagation import identity_propagation
from dune.perftool.ufl.transformations.reindexing import reindexing
from dune.perftool.ufl.modified_terminals import analyse_modified_argument, ModifiedArgument

from ufl.classes import Zero, Identity, Indexed, IntValue, MultiIndex
from ufl.core.multiindex import indices

from pytools import Record


class AccumulationTerm(Record):
    def __init__(self,
                 term,
                 argument,
                 indexmap={},
                 ):
        assert isinstance(argument, ModifiedArgument)
        Record.__init__(self,
                        term=term,
                        argument=argument,
                        indexmap=indexmap,
                        )


@ufl_transformation(name="accterms", extraction_lambda=lambda l: [at.term for at in l])
def split_into_accumulation_terms(expr, indexmap={}):
    ret = []

    # Extract a list of modified terminals for the test function
    # One accumulation instruction will be generated for each of these.
    test_args = extract_modified_arguments(expr, argnumber=0)

    # Extract a list of modified terminals for the ansatz function
    # in jacobian forms.
    all_jacobian_args = extract_modified_arguments(expr, argnumber=1)

    for test_arg in test_args:
        # Do this as a multi step replacement procedure to avoid UFL nagging
        # about too much stuff in reconstructing the expressions

        # 1) We first cut the expression to the relevant modified test_function
        # Build a replacement dictionary
        replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                     free_indices=ma.expr.ufl_free_indices,
                                     index_dimensions=ma.expr.ufl_index_dimensions)
                       for ma in test_args}
        replacement[test_arg.expr] = test_arg.expr
        replace_expr = replace_expression(expr, replacemap=replacement)

        # 2) Cut the test function itself from the expression
        indexmap = {}
        if test_arg.index:
            newi = indices(len(test_arg.index))
            identities = tuple(Indexed(Identity(2), MultiIndex((i,) + (j,))) for i, j in zip(newi, test_arg.index._indices))
            indexmap = {i: j for i, j in zip(test_arg.index._indices, newi)}
            from dune.perftool.ufl.flatoperators import construct_binary_operator
            from ufl.classes import Product
            replacement = {test_arg.expr: construct_binary_operator(identities, Product)}
            test_arg = analyse_modified_argument(reindexing(test_arg.expr, replacemap=indexmap))
        else:
            replacement = {test_arg.expr: IntValue(1)}
        replace_expr = replace_expression(replace_expr, replacemap=replacement)

        # 3) Collapse any identity nodes that may have been introduced by replacing vectors
        replace_expr = identity_propagation(replace_expr)

        # 4) Further split according to trial function in jacobian terms
        if all_jacobian_args:
            # Update the list!
            jac_args = extract_modified_arguments(replace_expr, argnumber=1)
            for jac_arg in jac_args:
                # TODO Some jacobian terms can be joined
                replacement = {ma.expr: Zero(shape=ma.expr.ufl_shape,
                                             free_indices=ma.expr.ufl_free_indices,
                                             index_dimensions=ma.expr.ufl_index_dimensions)
                               for ma in jac_args}
                replacement[jac_arg.expr] = jac_arg.expr
                jac_expr = replace_expression(replace_expr, replacemap=replacement)

                if not isinstance(jac_expr, Zero):
                    ret.append(AccumulationTerm(jac_expr, test_arg, indexmap))
        else:
            if not isinstance(replace_expr, Zero):
                ret.append(AccumulationTerm(replace_expr, test_arg, indexmap))

    return ret
